library ieee;
use ieee.std_logic_1164.all;
USE ieee.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;

-- boitier mmoire de 2**N de TM bits
-- adresse sur N bits
--
-- pour le projet prendre N=11, M=8 et L=20
--
-- auteur : Matthias Colin
-- version : 1.0 (3/05/2010)
--
-- Fonctionnement :
-- 
-- les resultats sont disponibles L cycles aprs rception
-- de la demande
--
-- le composant peut lire ou crire des donnes en rafale  la condition :
--	- le type d'operation ne change pas (ecriture ou lecture)
-- dans ce cas il suffit de maintenir le signal enable a true pdt la dure de l'opration
-- dans le cas contraire, il faut attendre la fin de l'opration en cours avant la suivante
-- dans le cas d'une mauvaise utilisation la mmoire se bloque et il faut faire un reset

entity main_memory is
generic	(
	-- paramtres de de la mmoire principale --
	N : integer range 1 to 64; 	-- taille de l'adresse => taille de la mmoire 2**N mots
	TM : integer range 0 to 8; 	-- taille d'un mot : 2**TM
	L : integer range 1 to 64	-- latence de la mmoire
);
port (
	reset : in std_logic; -- rinitialisation (niveau bas)
	clock : in std_logic; -- horloge (front montant)
	data_in : in std_logic_vector(1 to TM);		-- donne  crire
	data_out : out std_logic_vector(1 to TM);	-- donne  lire
	address : in std_logic_vector(N-1 downto 0); -- adresse de la donne  crire ou lire
	rw : in std_logic; -- 1 : read / 0 : write
	cs : in std_logic -- activation du boitier (niveau bas)
	
);
end main_memory;

architecture main_memory_arch of main_memory is
-- les tats de la mmoire
-- RDY : mmoire prte
-- WRAF : mmoire en cours d'ecriture (demande+traitement)
-- WFIN : mmoire en cours d'ecriture (traitement)
-- RRAF : mmoire en cours de lecture (demande+traitement)
-- RFIN : mmoire en cours de lecture (traitement)
-- ERR : mmoire bloque 
type t_etat is (RDY, WRAF, WFIN, RRAF, RFIN, ERR);
signal etat : t_etat;
	
-- le tableau des mots en mmoire
type t_memory is array(0 to 2**N-1) of std_logic_vector(1 to TM);
signal mots : t_memory;

-- dernires cases accdes (buffer pour les oprations en rafales)
-- 1. adresses :
type tabIndices is array(0 to L - 1) of integer range 0 to 2**N-1;
signal oldIndices : tabIndices;
-- 2. mots  crire
type t_buffer_mot is array(0 to L - 1) of std_logic_vector(1 to TM);		
signal buffer_mot : t_buffer_mot;

begin
	
	process(clock, reset) is
		-- index pour parcourir la memoire
		variable i : integer range 0 to 2**N-1; 
		-- compteurs
		variable cptO : integer range 0 to L-1; -- indice du prochain ordre en demande dans le buffer (plus rcent)
		variable cptT : integer range 0 to L-1; -- indice de la prochaine requte  traiter (plus ancienne)
		variable cpt : integer range 0 to L-1;	-- compteur pour la latence (cpt = L-1 => dbut taitement)	
	begin
		if reset='0' then
			-- initialisation de la memoire
			-- avec la valeur zro
			FOR i IN 0 to 2**11-1 LOOP
				mots(i) <= (others => '0');
			END LOOP;
			-- la mmoire est prte  tre lue ou crite
			etat <= RDY;
		elsif clock'event and clock='1' then
			-- traitement d'un front montant de l'horloge
			case etat is
				-- ETAT PRET INITIAL --
				when RDY =>
					if (cs = '0') then
						-- botier mmoire actif
						cptO := 0;
						cptT := 0;						
						cpt := 0;			
						-- sauvegarde de l'indice pour
						-- les lectures/ecritures en rafale
						i := conv_integer(address);
						oldIndices(cptO) <= i;						
						if rw = '1' then
						-- lecture --
							etat <= RRAF;
						else
						-- ecriture --
							-- sauvegarde du mot  crire
							buffer_mot(cptO) <= data_in;
							etat <= WRAF;
						end if;
					end if;
				-- ETAT ECRITURES EN RAFALES --	
				when WRAF=>
					-- RECEPTION REQUETES --
					if (cs = '0') then
					-- botier mmoire actif
						if rw = '1' then
						-- lecture : impossible juste aprs une criture --
							etat <= ERR;
						else
						-- criture --
							-- sauvegarde de la demande suivante
							cptO := (cptO + 1) mod L;
							i := conv_integer(address);						
							oldIndices(cptO) <= i;
							buffer_mot(cptO) <= data_in;							
						end if;
					else
					-- botier inactif (fin des demandes)
						etat <= WFIN;
					end if;
					
					-- TRAITEMENT REQUETES --
					if cpt = L - 1 then
						-- temps de latence atteint, on peut procder au traitement
						mots(oldIndices(cptT)) <= buffer_mot(cptT);
						cptT := (cptT + 1) mod L;
					else
						-- temps de latence non atteint, on attend un coup de plus					
						cpt := cpt + 1;
					end if;
				-- ETAT FINIR LES LECTURES EN COURS (TRAITEMENT UNIQUEMENT) --						
				when WFIN =>
					if (cs = '0') then
						-- temps d'attente non atteint avant demande suivante
						etat <= ERR;
					else
						if cpt = L - 1 then
						-- temps de latence atteint, on peut procder au traitement
							if cptO = cptT then
								-- fin des traitements
								etat <= RDY;
							end if;
							mots(oldIndices(cptT)) <= buffer_mot(cptT);
							cptT := (cptT + 1) mod L;
						else
						-- temps de latence non atteint, on attend un coup de plus					
							cpt := cpt + 1;
						end if;
					end if;
				-- ETAT LECTURES EN RAFALES --	
				when RRAF =>
					-- RECEPTION REQUETES --				
					if (cs = '0') then
					-- botier mmoire actif
						if rw = '1' then
						-- lecture --
							-- sauvegarde requete
							cptO := (cptO + 1) mod L;
							i := conv_integer(address);
							oldIndices(cptO) <= i;
						else
						-- criture : impossible juste aprs une lecture --
							etat <= ERR;
						end if;
					else
					-- botier inactif (fin des requetes)
						etat <= RFIN;
					end if;

					-- TRAITEMENT REQUETES --					
					if cpt = L - 1 then
						-- temps de latence atteint, on peut procder au traitement
						data_out <= mots(oldIndices(cptT));
						cptT := (cptT + 1) mod L;
					else
						-- temps de latence non atteint
						cpt := cpt + 1;
					end if;

				-- ETAT FINIR LES LECTURES EN COURS (TRAITEMENT UNIQUEMENT) --						
				when RFIN =>
					if (cs = '0') then
						-- temps d'attente non atteint avant demande suivante
						etat <= ERR;
					else
						if cpt = L - 1 then
						-- temps de latence atteint, on peut procder au traitement
							if cptO = cptT then
								-- fin des traitements
								etat <= RDY;
							end if;
							data_out <= mots(oldIndices(cptT));
							cptT := (cptT + 1) mod L;
						else
						-- temps de latence non atteint					
							cpt := cpt + 1;
						end if;
					end if;				
				-- ETAT ERREUR --	
				when ERR =>
					-- etat terminal : memoire plantee
					null;
			end case;
		end if;
	end process;
	
end main_memory_arch;	




